/*
** bootstrap.c -- This program loads the Linux/m68k kernel into an Amiga
**		  and launches it.
**
** Copyright 1993,1994 by Hamish Macdonald, Greg Harp
**
** Modified 11-May-94 by Geert Uytterhoeven
**			(Geert.Uytterhoeven@cs.kuleuven.ac.be)
**     - A3640 MapROM check
** Modified 31-May-94 by Geert Uytterhoeven
**     - Memory thrash problem solved
** Modified 07-March-95 by Geert Uytterhoeven
**     - Memory block sizes are rounded to a multiple of 256K instead of 1M
**	 This _requires_ >0.9pl5 to work!
**	 (unless all block sizes are multiples of 1M :-)
** Modified 11-July-95 by Andreas Schwab
**     - Support for ELF kernel (untested!)
** Modified 10-Jan-96 by Geert Uytterhoeven
**     - The real Linux/m68k boot code moved to linuxboot.[ch]
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
*/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/file.h>
#include <sys/types.h>
#include <unistd.h>

/* required Linux/m68k include files */
#include <linux/a.out.h>
#include <linux/elf.h>
#include <asm/bootinfo.h>
#include <asm/page.h>

/* Amiga bootstrap include files */
#include "linuxboot.h"
#include "bootstrap.h"


/* Exec Base */
extern struct ExecBase *SysBase;

extern char *optarg;

struct ExpansionBase *ExpansionBase;
struct GfxBase *GfxBase;

static char *memfile = NULL;

static int model = AMI_UNKNOWN;


static void usage(void)
{
    fprintf(stderr, "Usage:\n"
		    "\tbootstrap [-d] [-v] [-k kernel_executable] "
		    " [-r ramdisk_file] [-m memory_file] [-t amiga_model]"
		    " [option...]\n");
    exit (EXIT_FAILURE);
}


int main(int argc, char *argv[])
{
    int ch, i;
    int debugflag = 0, keep_video = 0;
    char *kernel_name = "vmlinux";
    char *ramdisk_name = NULL;
    char commandline[CL_SIZE] = "";

    while ((ch = getopt(argc, argv, "dvk:r:m:t:")) != EOF)
	switch (ch) {
	    case 'd':
		debugflag = 1;
		break;
	    case 'v':
		keep_video = 1;
		break;
	    case 'k':
		kernel_name = optarg;
		break;
	    case 'r':
		ramdisk_name = optarg;
		break;
	    case 'm':
		memfile = optarg;
		break;
	    case 't':
		model = atoi(optarg);
		break;
	    case '?':
	    default:
		usage();
	}
    argc -= optind;
    argv += optind;

    SysBase = *(struct ExecBase **)4;

    /* open Expansion Library */
    ExpansionBase = (struct ExpansionBase *)OpenLibrary("expansion.library",
							36);
    if (!ExpansionBase) {
	fputs("Unable to open expansion.library V36 or greater!  Aborting...\n",
	      stderr);
	exit(EXIT_FAILURE);
   }

    /* open Graphics Library */
    GfxBase = (struct GfxBase *)OpenLibrary ("graphics.library", 0);
    if (!GfxBase) {
	fputs("Unable to open graphics.library!  Aborting...\n", stderr);
	exit(EXIT_FAILURE);
    }

    /*
     *	Join command line options
     */
    i = 0;
    while (argc--) {
	if ((i+strlen(*argv)+1) < CL_SIZE) {
	    i += strlen(*argv) + 1;
	    if (commandline[0])
		strcat(commandline, " ");
	    strcat(commandline, *argv++);
	}
    }

    /* Do The Right Stuff */
    BootLinux(kernel_name, ramdisk_name, commandline, debugflag, keep_video);

    CloseLibrary((struct Library *)GfxBase);
    CloseLibrary((struct Library *)ExpansionBase);

    /* if we ever get here, something went wrong */
    exit(EXIT_FAILURE);
}

/* For communication from `getopt' to the caller.
   When `getopt' finds an option that takes an argument,
   the argument value is returned here.
   Also, when `ordering' is RETURN_IN_ORDER,
   each non-option ARGV-element is returned here.  */

char *optarg = 0;

/* Index in ARGV of the next element to be scanned.
   This is used for communication to and from the caller
   and for communication between successive calls to `getopt'.

   On entry to `getopt', zero means this is the first call; initialize.

   When `getopt' returns EOF, this is the index of the first of the
   non-option elements that the caller should itself scan.

   Otherwise, `optind' communicates from one call to the next
   how much of ARGV has been scanned so far.  */

int optind = 0;

/* The next char to be scanned in the option-element
   in which the last option character we returned was found.
   This allows us to pick up the scan where we left off.

   If this is zero, or a null string, it means resume the scan
   by advancing to the next ARGV-element.  */

static char *nextchar;

/* Callers store zero here to inhibit the error message
   for unrecognized options.  */

int opterr = 1;

/* Describe how to deal with options that follow non-option ARGV-elements.

   If the caller did not specify anything,
   the default is REQUIRE_ORDER if the environment variable
   _POSIX_OPTION_ORDER is defined, PERMUTE otherwise.

   REQUIRE_ORDER means don't recognize them as options.
   Stop option processing when the first non-option is seen.
   This is what Unix does.

   PERMUTE is the default.  We permute the contents of ARGV as we scan,
   so that eventually all the options are at the end.  This allows options
   to be given in any order, even with programs that were not written to
   expect this.

   RETURN_IN_ORDER is an option available to programs that were written
   to expect options and other ARGV-elements in any order and that care about
   the ordering of the two.  We describe each non-option ARGV-element
   as if it were the argument of an option with character code one.
   Using `-' as the first character of the list of option characters
   requests this mode of operation.

   The special argument `--' forces an end of option-scanning regardless
   of the value of `ordering'.  In the case of RETURN_IN_ORDER, only
   `--' can cause `getopt' to return EOF with `optind' != ARGC.  */

static enum { REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER } ordering;

/* Describe the long-named options requested by the application.
   _GETOPT_LONG_OPTIONS is a vector of `struct option' terminated by an
   element containing a name which is zero.
   The field `has_arg' is 1 if the option takes an argument, 
   2 if it takes an optional argument.  */

struct option
{
  char *name;
  int has_arg;
  int *flag;
  int val;
};

const struct option *_getopt_long_options;

int _getopt_long_only = 0;

/* Index in _GETOPT_LONG_OPTIONS of the long-named option actually found.
   Only valid when a long-named option was found. */

int option_index;

/* Handle permutation of arguments.  */

/* Describe the part of ARGV that contains non-options that have
   been skipped.  `first_nonopt' is the index in ARGV of the first of them;
   `last_nonopt' is the index after the last of them.  */

static int first_nonopt;
static int last_nonopt;

/* Exchange two adjacent subsequences of ARGV.
   One subsequence is elements [first_nonopt,last_nonopt)
    which contains all the non-options that have been skipped so far.
   The other is elements [last_nonopt,optind), which contains all
    the options processed since those non-options were skipped.

   `first_nonopt' and `last_nonopt' are relocated so that they describe
    the new indices of the non-options in ARGV after they are moved.  */

static void
exchange (char * const *argv)
{
  int nonopts_size
    = (last_nonopt - first_nonopt) * sizeof (char *);
  char **temp = (char **) alloca (nonopts_size);

  /* Interchange the two blocks of data in argv.  */

  bcopy (&argv[first_nonopt], temp, nonopts_size);
  bcopy (&argv[last_nonopt], (char **)&argv[first_nonopt],
         (optind - last_nonopt) * sizeof (char *));
  bcopy (temp, (char **)&argv[first_nonopt + optind - last_nonopt],
         nonopts_size);

  /* Update records for the slots the non-options now occupy.  */

  first_nonopt += (optind - last_nonopt);
  last_nonopt = optind;
}

/* Scan elements of ARGV (whose length is ARGC) for option characters
   given in OPTSTRING.

   If an element of ARGV starts with '-', and is not exactly "-" or "--",
   then it is an option element.  The characters of this element
   (aside from the initial '-') are option characters.  If `getopt'
   is called repeatedly, it returns successively each of the option characters
   from each of the option elements.

   If `getopt' finds another option character, it returns that character,
   updating `optind' and `nextchar' so that the next call to `getopt' can
   resume the scan with the following option character or ARGV-element.

   If there are no more option characters, `getopt' returns `EOF'.
   Then `optind' is the index in ARGV of the first ARGV-element
   that is not an option.  (The ARGV-elements have been permuted
   so that those that are not options now come last.)

   OPTSTRING is a string containing the legitimate option characters.
   If an option character is seen that is not listed in OPTSTRING,
   return '?' after printing an error message.  If you set `opterr' to
   zero, the error message is suppressed but we still return '?'.

   If a char in OPTSTRING is followed by a colon, that means it wants an arg,
   so the following text in the same ARGV-element, or the text of the following
   ARGV-element, is returned in `optarg'.  Two colons mean an option that
   wants an optional arg; if there is text in the current ARGV-element,
   it is returned in `optarg', otherwise `optarg' is set to zero.

   If OPTSTRING starts with `-', it requests a different method of handling the
   non-option ARGV-elements.  See the comments about RETURN_IN_ORDER, above.

   Long-named options begin with `+' instead of `-'.
   Their names may be abbreviated as long as the abbreviation is unique
   or is an exact match for some defined option.  If they have an
   argument, it follows the option name in the same ARGV-element, separated
   from the option name by a `=', or else the in next ARGV-element.
   `getopt' returns 0 when it finds a long-named option.  */

int
getopt (argc, argv, optstring)
     int argc;
     char * const *argv;
     const char *optstring;
{
  optarg = 0;

  /* Initialize the internal data when the first call is made.
     Start processing options with ARGV-element 1 (since ARGV-element 0
     is the program name); the sequence of previously skipped
     non-option ARGV-elements is empty.  */

  if (optind == 0)
    {
      first_nonopt = last_nonopt = optind = 1;

      nextchar = 0;

      /* Determine how to handle the ordering of options and nonoptions.  */

      if (optstring[0] == '-')
        ordering = RETURN_IN_ORDER;
      else if (getenv ("_POSIX_OPTION_ORDER") != 0)
        ordering = REQUIRE_ORDER;
      else
        ordering = PERMUTE;
    }

  if (nextchar == 0 || *nextchar == 0)
    {
      if (ordering == PERMUTE)
        {
          /* If we have just processed some options following some non-options,
             exchange them so that the options come first.  */

          if (first_nonopt != last_nonopt && last_nonopt != optind)
            exchange (argv);
          else if (last_nonopt != optind)
            first_nonopt = optind;

          /* Now skip any additional non-options
             and extend the range of non-options previously skipped.  */

          while (optind < argc
                 && (argv[optind][0] != '-'
                     || argv[optind][1] == 0)
                 && (_getopt_long_options == 0
                     || argv[optind][0] != '+'
                     || argv[optind][1] == 0))
            optind++;
          last_nonopt = optind;
        }

      /* Special ARGV-element `--' means premature end of options.
         Skip it like a null option,
         then exchange with previous non-options as if it were an option,
         then skip everything else like a non-option.  */

      if (optind != argc && !strcmp (argv[optind], "--"))
        {
          optind++;

          if (first_nonopt != last_nonopt && last_nonopt != optind)
            exchange (argv);
          else if (first_nonopt == last_nonopt)
            first_nonopt = optind;
          last_nonopt = argc;

          optind = argc;
        }

      /* If we have done all the ARGV-elements, stop the scan
         and back over any non-options that we skipped and permuted.  */

      if (optind == argc)
        {
          /* Set the next-arg-index to point at the non-options
             that we previously skipped, so the caller will digest them.  */
          if (first_nonopt != last_nonopt)
            optind = first_nonopt;
          return EOF;
        }
         
      /* If we have come to a non-option and did not permute it,
         either stop the scan or describe it to the caller and pass it by.  */

      if ((argv[optind][0] != '-' || argv[optind][1] == 0)
          && (_getopt_long_options == 0
              || argv[optind][0] != '+' || argv[optind][1] == 0))
        {
          if (ordering == REQUIRE_ORDER)
            return EOF;
          optarg = argv[optind++];
          return 1;
        }

      /* We have found another option-ARGV-element.
         Start decoding its characters.  */

      nextchar = argv[optind] + 1;
    }

  if (_getopt_long_options != 0
      && (argv[optind][0] == '+'
          || (_getopt_long_only && argv[optind][0] == '-'))
      )
    {
      const struct option *p;
      char *s = nextchar;
      int exact = 0;
      int ambig = 0;
      const struct option *pfound = 0;
      int indfound = 0;

      while (*s && *s != '=') s++;

      /* Test all options for either exact match or abbreviated matches.  */
      for (p = _getopt_long_options, option_index = 0; p->name; 
           p++, option_index++)
        if (!strncmp (p->name, nextchar, s - nextchar))
          {
            if (s - nextchar == strlen (p->name))
              {
                /* Exact match found.  */
                pfound = p;
                indfound = option_index;
                exact = 1;
                break;
              }
            else if (pfound == 0)
              {
                /* First nonexact match found.  */
                pfound = p;
                indfound = option_index;
              }
            else
              /* Second nonexact match found.  */
              ambig = 1;
          }

      if (ambig && !exact)
        {
          fprintf (stderr, "%s: option `%s' is ambiguous\n",
                   argv[0], argv[optind]);
          nextchar += strlen (nextchar);                           
          return '?';
        }

      if (pfound != 0)
        {
          option_index = indfound;
          optind++;
          if (*s)
            {
              if (pfound->has_arg > 0)
                optarg = s + 1;
              else
                {
                  fprintf (stderr,
                           "%s: option `%c%s' doesn't allow an argument\n",
                           argv[0], argv[optind - 1][0], pfound->name);
                  nextchar += strlen (nextchar);                           
                  return '?';
                }
            }
          else if (pfound->has_arg == 1)
            {
              if (optind < argc)
                optarg = argv[optind++];
              else
                {
                  fprintf (stderr, "%s: option `%s' requires an argument\n",
                           argv[0], argv[optind - 1]);
                  nextchar += strlen (nextchar);                   
                  return '?';
                }
            }
          nextchar += strlen (nextchar);
          if (pfound->flag)
            *(pfound->flag) = pfound->val;
          return 0;
        }
      /* Can't find it as a long option.  If this is getopt_long_only,
         and the option starts with '-' and is a valid short
         option, then interpret it as a short option.  Otherwise it's
         an error.  */
      if (_getopt_long_only == 0 || argv[optind][0] == '+' ||
          index (optstring, *nextchar) == 0)
        {
          if (opterr != 0)
            fprintf (stderr, "%s: unrecognized option `%c%s'\n",
                     argv[0], argv[optind][0], nextchar);
          nextchar += strlen (nextchar);                   
          return '?';
        }
    }
 
  /* Look at and handle the next option-character.  */

  {
    char c = *nextchar++;
    char *temp = index (optstring, c);

    /* Increment `optind' when we start to process its last character.  */
    if (*nextchar == 0)
      optind++;

    if (temp == 0 || c == ':')
      {
        if (opterr != 0)
          {
            if (c < 040 || c >= 0177)
              fprintf (stderr, "%s: unrecognized option, character code 0%o\n",
                       argv[0], c);
            else
              fprintf (stderr, "%s: unrecognized option `-%c'\n",
                       argv[0], c);
          }
        return '?';
      }
    if (temp[1] == ':')
      {
        if (temp[2] == ':')
          {
            /* This is an option that accepts an argument optionally.  */
            if (*nextchar != 0)
              {
                optarg = nextchar;
                optind++;
              }
            else
              optarg = 0;
            nextchar = 0;
          }
        else
          {
            /* This is an option that requires an argument.  */
            if (*nextchar != 0)
              {
                optarg = nextchar;
                /* If we end this ARGV-element by taking the rest as an arg,
                   we must advance to the next element now.  */
                optind++;
              }
            else if (optind == argc)
              {
                if (opterr != 0)
                  fprintf (stderr, "%s: option `-%c' requires an argument\n",
                           argv[0], c);
                c = '?';
              }
            else
              /* We already incremented `optind' once;
                 increment it again when taking next ARGV-elt as argument.  */
              optarg = argv[optind++];
            nextchar = 0;
          }
      }
    return c;
  }
}


void Puts(char *str)
{
   fputs(str, stderr);
}


long GetChar(void)
{
   return(getchar());
}


void PutChar(char c)
{
   fputc(c, stderr);
}


void Printf(char *fmt, ...)
{
   va_list args;

   va_start(args, fmt);
   vfprintf(stderr, fmt, args);
   va_end(args);
}


int Open(u_long key)
{
   return(open((char *)key, O_RDONLY));
}


int Seek(int fd, int offset)
{
   return(lseek(fd, offset, SEEK_SET));
}


int Read(int fd, char *buf, int count)
{
   return(read(fd, buf, count));
}


void Close(int fd)
{
   close(fd);
}

int FileSize(u_long key)
{
   int fd, size = -1;

   if ((fd = open((char *)key, O_RDONLY)) != -1) {
      size = seek(rfd, 0, SEEK_END);
      close(fd);
   }
   return(size);
}

void Sleep(u_long micros)
{
   struct MsgPort *TimerPort;
   struct timerequest *TimerRequest;

   if (!(TimerPort = CreateMsgPort()))
      goto noport;

   if (!(TimerRequest = CreateIORequest(TimerPort,
                                        sizeof(struct timerequest))))
      goto noioreq;

   if (OpenDevice("timer.device", UNIT_VBLANK, (struct IORequest *)TimerRequest,
                  0))
      goto nodev;

   TimerRequest->io_Command = TR_ADDREQUEST;
   TimerRequest->io_Flags = IOF_QUICK;
   TimerRequest->tv_secs = micros/1000000;
   TimerRequest->tv_micro = micros%1000000;
   DoIO((struct IORequest *)TimerRequest);

   CloseDevice((struct IORequest *)TimerRequest);
nodev:
   DeleteIORequest(TimerRequest);
noioreq:
   DeleteMsgPort(TimerPort);
noport:
   return;
}


int ModifyBootInfo(struct bootinfo *bi)
{
   /*
    * if we have a memory file, read the memory information from it
    */
   if (memfile) {
      FILE *fp;
      int i;

      if ((fp = fopen(memfile, "r")) == NULL) {
         perror("open memory file");
         fprintf(stderr, "Cannot open memory file %s\n", memfile);
         return(FALSE);
      }

      if (fscanf(fp, "%lu", &bi->bi_amiga.chip_size) != 1) {
         fprintf(stderr, "memory file does not contain chip memory size\n");
         fclose(fp);
         return(FALSE);
      }
                
      for (i = 0; i < NUM_MEMINFO; i++) {
         if (fscanf(fp, "%lx %lu", &bi->memory[i].addr,
                    &bi->memory[i].size) != 2)
            break;
      }

      fclose(fp);

      if (i != bi->num_memory && i > 0)
         bi->num_memory = i;
   }

   /*
    * change the Amiga model, if necessary
    */
   if (model != AMI_UNKNOWN)
      bi->bi_amiga.model = model;

   return(TRUE);
}
