/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *		       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *  --------------------------------------------------------------------------
 *
 *  The original idea for this program comes from the Linux distribution:
 *
 *	LILO - Generic Boot Loader for Linux ("LInux LOader")
 *	       by Werner Almesberger
 *
 *  --------------------------------------------------------------------------
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file README.legal in the main directory of the
 *  Linux/m68k distribution for more details.
 */


%{

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include "lilo.h"


extern int yylex(void);
extern void yyerror(char *s);
extern int line;


static struct BootRecord BootRecord;


static void ClearBootRecord()
{
    memset(&BootRecord, 0, sizeof(BootRecord));
}


static void Redefinition(char *name)
{
    printf("%s:%d: Redefinition of `%s'\n", ConfigFile, line, name);
}


static u_long *CopyLong(u_long data)
{
    u_long *p;

    if (!(p = malloc(sizeof(u_long))))
	Die("No memory\n");
    *p = data;
    return(p);
}


static u_long *CopyLong2(u_long data1, u_long data2)
{
    u_long *p;

    if (!(p = malloc(2*sizeof(u_long))))
	Die("No memory\n");
    p[0] = data1;
    p[1] = data2;
    return(p);
}

%}

%start file

%token HEADER ENDHEADER BOOTREC ENDREC BOOTDEV ALTDEV DEFAULT AUTO TIMEOUT AUX
       BAUD PASSWORD DEBUG PROMPT TYPE IMAGE ARGS CHIPRAM FASTRAM MODEL ALIAS
       RAMDISK STRING NUMBER BOOLEAN BOOTOS AMIMODEL

%%

file	  : header bootrecs
	  ;

header	  : HEADER hdropts ENDHEADER
	    {
		if (!Options.BootDevice)
		    yyerror("No boot device specified");
		ClearBootRecord();
	    }
	  ;

hdropts	  : /* empty */
	  | hdropts bootdev
	  | hdropts altdev
	  | hdropts default
	  | hdropts auto
	  | hdropts timeout
	  | hdropts aux
	  | hdropts baud
	  | hdropts mrpasswd
	  | hdropts debug
	  | hdropts prompt
	  ;

bootdev	  : BOOTDEV STRING
	    {
		if (Options.BootDevice)
		    Redefinition((char *)$1);
		Options.BootDevice = (char *)$2;
	    }
	  ;

altdev    : ALTDEV STRING NUMBER
	    {
		if (Options.AltDeviceName)
		    Redefinition((char *)$1);
		Options.AltDeviceName = (char *)$2;
		Options.AltDeviceUnit = CopyLong($3);
		if (strlen(Options.AltDeviceName) >= MAXALTDEVSIZE)
		    yyerror("Alternate boot device name is too long");
	    }
	  ;

default   : DEFAULT STRING
	    {
		if (Options.Boot.Default)
		    Redefinition((char *)$1);
		Options.Boot.Default = (char *)$2;
	    }
	  ;

auto      : AUTO BOOLEAN
	    {
		if (Options.Boot.Auto)
		    Redefinition((char *)$1);
		Options.Boot.Auto = CopyLong($2);
	    }
	  ;

timeout   : TIMEOUT NUMBER
	    {
		if (Options.Boot.TimeOut)
		    Redefinition((char *)$1);
		Options.Boot.TimeOut = CopyLong($2);
	    }
	  ;

aux       : AUX BOOLEAN
	    {
		if (Options.Boot.Aux)
		    Redefinition((char *)$1);
		Options.Boot.Aux = CopyLong($2);
	    }
	  ;

baud      : BAUD NUMBER
	    {
		if (Options.Boot.Baud)
		    Redefinition((char *)$1);
		Options.Boot.Baud = CopyLong($2);
	    }
	  ;

mrpasswd  : PASSWORD STRING
	    {
		if (Options.Boot.MasterPassword)
		    Redefinition((char *)$1);
		Options.Boot.MasterPassword = (char *)$2;
	    }
	  ;

debug     : DEBUG BOOLEAN
	    {
		if (Options.Boot.Debug)
		    Redefinition((char *)$1);
		Options.Boot.Debug = CopyLong($2);
	    }
	  ;

prompt	  : PROMPT STRING
	    {
		if (Options.Boot.Prompt)
		    Redefinition((char *)$1);
		Options.Boot.Prompt = (char *)$2;
	    }
	  ;

bootrecs  : bootrec
	  | bootrecs bootrec
	  ;

bootrec   : BOOTREC STRING bootopts ENDREC
	    {
		BootRecord.Label = (char *)$2;
		if (BootRecord.OSType && *BootRecord.OSType == BOS_AMIGA) {
		    if (BootRecord.ImagePath)
			yyerror("You can't specify an image for AmigaOS");
		    if (BootRecord.Args)
			yyerror("You can't specify arguments for AmigaOS");
		    if (BootRecord.RamdiskPath)
			yyerror("You can't specify a ramdisk file for AmigaOS");
		} else if (!BootRecord.ImagePath)
		    yyerror("No boot image specified");
		if (!AddBootRecord(&BootRecord))
		    yyerror("Duplicate boot image label or alias");
		ClearBootRecord();
	    }
	  ;

bootopts  : /* empty */
	  | bootopts type
	  | bootopts image
	  | bootopts args
	  | bootopts passwd
	  | bootopts chipram
	  | bootopts fastram
	  | bootopts model
	  | bootopts modelnum
	  | bootopts alias
	  | bootopts ramdisk
	  ;

type      : TYPE BOOTOS
	    {
		if (BootRecord.OSType)
		    Redefinition((char *)$1);
		BootRecord.OSType = CopyLong($2);
	    }
	  ;

image     : IMAGE STRING
	    {
		if (BootRecord.ImagePath)
		    Redefinition((char *)$1);
		BootRecord.ImagePath = (char *)$2;
	    }
	  ;

args      : ARGS STRING
	    {
		if (BootRecord.Args)
		    Redefinition((char *)$1);
		BootRecord.Args = (char *)$2;
		if (strlen(BootRecord.Args) >= CL_SIZE)
		    yyerror("Boot argument is too long");
	    }
	  ;

passwd    : PASSWORD STRING
	    {
		if (BootRecord.Password)
		    Redefinition((char *)$1);
		BootRecord.Password = (char *)$2;
	    }
	  ;

chipram   : CHIPRAM NUMBER
	    {
		if (BootRecord.ChipSize)
		    Redefinition((char *)$1);
		BootRecord.ChipSize = CopyLong($2);
	    }
	  ;

fastram   : FASTRAM NUMBER NUMBER
	    {
		int i;

		for (i = 0; i < arraysize(BootRecord.FastChunks); i++)
		    if (!BootRecord.FastChunks[i])
			break;
		if (i == arraysize(BootRecord.FastChunks))
		    yyerror("Too many Fast RAM Chunks");
		BootRecord.FastChunks[i] = CopyLong2($2, $3);
	    }

model     : MODEL AMIMODEL
	    {
		if (BootRecord.Model)
		    Redefinition((char *)$1);
		BootRecord.Model = CopyLong($2);
	    }
	  ;

modelnum  : MODEL NUMBER
	    {
		if (BootRecord.Model)
		    Redefinition((char *)$1);
		BootRecord.Model = CopyLong($2);
	    }
	  ;

alias	  : ALIAS STRING
	    {
		if (BootRecord.Alias)
		    Redefinition((char *)$1);
		BootRecord.Alias = (char *)$2;
	    }
	  ;

ramdisk	  : RAMDISK STRING
	    {
		if (BootRecord.RamdiskPath)
		    Redefinition((char *)$1);
		BootRecord.RamdiskPath = (char *)$2;
	    }
	  ;

%%
