#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/fcntl.h>

#include "lilo.h"


void Die(const char *fmt, ...)
{
    va_list ap;

    fflush(stdout);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}


int main(int argc, char *argv[])
{
    int fh, size = 0;
    u_char *data;
    struct TagRecord *tr;
    u_long offset = 0;

    if (argc < 2) {
	printf("Usage: %s <mapfile>\n", argv[0]);
	return(10);
    }

    if ((fh = open(argv[1], O_RDONLY)) == -1)
	Die("open %s: %s\n", argv[1], strerror(errno));
    if ((size = lseek(fh, 0, SEEK_END)) == -1)
	Die("lseek %s: %s\n", argv[1], strerror(errno));
    if (!(data = malloc(size)))
	Die("No memory\n");
    if (lseek(fh, 0, SEEK_SET) == -1)
	Die("lseek %s: %s\n", argv[1], strerror(errno));
    if (read(fh, data, size) != size)
	Die("read %s: %s\n", argv[1], strerror(errno));
    close(fh);

    while (1) {
	tr = (struct TagRecord *)&data[offset];
	printf("0x%08lx: tag = 0x%08lx, size = 0x%08lx\n", offset,
	       tr->Tag, tr->Size);
	switch (tr->Tag) {
	    case TAG_LILO:
		printf("\t    TAG_LILO: File Identification\n");
		break;
	    case TAG_HEADER:
		printf("\t    TAG_HEADER: Start of Boot Options\n");
		break;
	    case TAG_HEADER_END:
		printf("\t    TAG_HEADER_END: End of Boot Options\n");
		break;
	    case TAG_DEFAULT:
		printf("\t    TAG_DEFAULT: Default Boot Record = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_AUTO:
		printf("\t    TAG_AUTO: Auto = %ld\n", tr->Data[0]);
		break;
	    case TAG_TIME_OUT:
		printf("\t    TAG_TIME_OUT: Time Out = %ld seconds\n",
		       tr->Data[0]);
		break;
	    case TAG_AUX:
		printf("\t    TAG_AUX: Aux = %ld\n", tr->Data[0]);
		break;
	    case TAG_BAUD:
		printf("\t    TAG_BAUD: Baud = %ld\n", tr->Data[0]);
		break;
	    case TAG_MASTER_PASSWORD:
		printf("\t    TAG_MASTER_PASSWORD: Master Password = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_DEBUG:
		printf("\t    TAG_DEBUG: Debug = %ld\n", tr->Data[0]);
		break;
	    case TAG_PROMPT:
		printf("\t    TAG_PROMPT: Boot Prompt = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_BOOT_RECORD:
		printf("\t    TAG_BOOT_RECORD: Start of Boot Record `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_BOOT_RECORD_END:
		printf("\t    TAG_BOOT_RECORD_END: End of Boot Record\n");
		break;
	    case TAG_ALIAS:
		printf("\t    TAG_ALIAS: Alias Label = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_OS_TYPE:
		printf("\t    TAG_OS_TYPE: Operating System type = %ld\n",
		       tr->Data[0]);
		break;
	    case TAG_IMAGE_PATH:
		printf("\t    TAG_IMAGE_PATH: Boot Image Path = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_IMAGE_VECTOR:
		printf("\t    TAG_IMAGE_VECTOR: 0x%08lx Bytes\n",
		       tr->Data[0]);
		break;
	    case TAG_ARGUMENTS:
		printf("\t    TAG_ARGUMENTS: Boot Arguments = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_PASSWORD:
		printf("\t    TAG_PASSWORD: Boot Record Password = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_CHIP_RAM_SIZE:
		printf("\t    TAG_CHIP_RAM_SIZE: 0x%08lx bytes of Chip RAM\n",
		       tr->Data[0]);
		break;
	    case TAG_FAST_RAM_CHUNK:
		printf("\t    TAG_FAST_RAM_CHUNK: 0x%08lx bytes at 0x%08lx\n",
		       tr->Data[1], tr->Data[0]);
		break;
	    case TAG_MODEL:
		printf("\t    TAG_MODEL: Model = %ld\n", tr->Data[0]);
		break;
	    case TAG_RAMDISK_PATH:
		printf("\t    TAG_RAMDISK_PATH: Ramdisk Path = `%s'\n",
		       (char *)tr->Data);
		break;
	    case TAG_RAMDISK_VECTOR:
		printf("\t    TAG_RAMDISK_VECTOR: 0x%08lx Bytes\n",
		       tr->Data[0]);
		break;
	    case TAG_EOF:
		printf("\t    TAG_EOF: End of File\n");
		break;
	    default:
		printf("\t    <UNKNOWN>\n");
		break;
	}
	offset += sizeof(struct TagRecord)+tr->Size;
	if (offset > size) {
	    puts("Warning: Beyond end of file");
	    break;
	}
	if (tr->Tag == TAG_EOF)
	    break;
    }
    if (offset < size)
	printf("Warning: Left %ld bytes\n", size-offset);
    return(0);
}
