/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *		       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *  --------------------------------------------------------------------------
 *
 *  The original idea for this program comes from the Linux distribution:
 *
 *	LILO - Generic Boot Loader for Linux ("LInux LOader")
 *	       by Werner Almesberger
 *
 *  --------------------------------------------------------------------------
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file README.legal in the main directory of the
 *  Linux/m68k distribution for more details.
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/hdreg.h>

#include "lilo.h"


    /*
     *	Command Line Options
     */

char *ProgramName = NULL;
int Verbose = 0;
int Install = 1;
int Force = 0;
int DoChecksum = 0;
char *DosType = NULL;
char *SaveBootBlock = NULL;
char *RestoreBootBlock = NULL;
char *Device = NULL;
char *Root = NULL;


    /*
     *	Boot Code Definitions
     */

extern u_char _BootCodeStart[];
extern u_char _BootCodeEnd[];
extern u_char _LoaderVectorStart[];
extern u_char _LoaderVectorEnd[];
extern u_char _AltDeviceName[];
extern u_char _AltDeviceUnit[];


    /*
     *	Valid DosTypes
     */

struct DosTypeEntry {
    u_long ID;
    char *ASCII;
    char *Description;
};

struct DosTypeEntry DosTypeEntries[] = {
    { ID_BOOT_DISK, "BOOU", "Generic boot disk" },
    { ID_OFS_DISK, "DOS0", "OFS disk" },
    { ID_FFS_DISK, "DOS1", "FFS disk" },
    { ID_OFS_INTL_DISK, "DOS2", "OFS INTL disk" },
    { ID_FFS_INTL_DISK, "DOS3", "FFS INTL disk" },
    { ID_OFS_DC_DISK, "DOS4", "OFS DC disk" },
    { ID_FFS_DC_DISK, "DOS5", "FFS DC disk" },
    { ID_MUFS_DISK, "muFS", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DISK, "muF0", "MultiUser OFS disk" },
    { ID_MU_FFS_DISK, "muF1", "MultiUser FFS disk" },
    { ID_MU_OFS_INTL_DISK, "muF2", "MultiUser OFS INTL disk" },
    { ID_MU_FFS_INTL_DISK, "muF3", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DC_DISK, "muF4", "MultiUser OFS DC disk" },
    { ID_MU_FFS_DC_DISK, "muF5", "MultiUser FFS DC disk" },
    { 0, NULL, NULL }
};


    /*
     *	Lilo Data
     */

struct BootBlock BootBlock;
struct BootRecord *BootRecords = NULL;

int MapFileFH;
u_long *MapVector;
int MapNumBlocks;

char *LoaderData;
int LoaderSize;
u_long *LoaderVector;
int LoaderNumBlocks;
u_long DosTypeID;

dev_t BootDevice;
struct stat info;
DIR *dir;
struct dirent *dirent;


    /*
     *	Configuration File Parsing
     */

struct Options Options = {
    NULL, NULL, NULL, {
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
    }
};


    /*
     *	Files
     */

char *ConfigFile, *MapFile, *LoaderTemplate, *LoaderFile;


    /*
     *	Function Prototypes
     */

void Die(const char *fmt, ...) __attribute__ ((noreturn));
int AddBootRecord(struct BootRecord *record);
int main(int argc, char *argv[]);

static void ReadBootBlock(char *name);
static void WriteBootBlock(char *name);
static int ReadFile(char *name, void **data);
static void WriteFile(char *name, void *data, int size);
static void ReadConfigFile(void);
static int EqualStrings(const char *s1, const char *s2);
static void CreateMapFile(void);
static void WriteTagData(u_long tag, void *data, u_long size);
static void WriteTagString(u_long tag, char *s);
static u_long CalcChecksum(void);
static void Geometry(char *name, dev_t *device, u_long *start, u_long *blksize);
static u_long *CreateVector(char *name, int *numblocks);
static void PatchLoader(void);
static void CreateBootBlock(void);
static void SetDosType(void);
static void FixChecksum(void);
static void Usage(void) __attribute__ ((noreturn));
static char *CreateFileName(const char *name);


    /*
     *	Print an Error Message and Exit
     */

void Die(const char *fmt, ...)
{
    va_list ap;

    fflush(stdout);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}


    /*
     *	Read a Boot Block from a Device or File
     */

static void ReadBootBlock(char *name)
{
    int fh;

    if ((fh = open(name, O_RDONLY)) == -1)
	Die("open %s: %s\n", name, strerror(errno));
    if (read(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Die("read %s: %s\n", name, strerror(errno));
    close(fh);
}


    /*
     *	Write a Boot Block to a Device
     */

static void WriteBootBlock(char *name)
{
    int fh;

    if ((fh = open(name, O_RDWR)) == -1)
	Die("open %s: %s\n", name, strerror(errno));
    if (write(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Die("write %s: %s\n", name, strerror(errno));
    close(fh);
}


    /*
     *	Read a File
     */

static int ReadFile(char *name, void **data)
{
    int fh, size = 0;

    if ((fh = open(name, O_RDONLY)) == -1)
	Die("open %s: %s\n", name, strerror(errno));
    if ((size = lseek(fh, 0, SEEK_END)) == -1)
	Die("lseek %s: %s\n", name, strerror(errno));
    if (!(*data = malloc(size)))
	Die("No memory\n");
    if (lseek(fh, 0, SEEK_SET) == -1)
	Die("lseek %s: %s\n", name, strerror(errno));
    if (read(fh, *data, size) != size)
	Die("read %s: %s\n", name, strerror(errno));
    close(fh);
    return(size);
}


    /*
     *	Write a File
     */

static void WriteFile(char *name, void *data, int size)
{
    int fh;

    if ((fh = open(name, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
	Die("open %s: %s\n", name, strerror(errno));
    if (write(fh, data, size) != size)
	Die("write %s: %s\n", name, strerror(errno));
    close(fh);
}


    /*
     *	Read the Configuration File
     */

static void ReadConfigFile(void)
{
    if (!(yyin = fopen(ConfigFile, "r")))
	Die("fopen %s: %s\n", ConfigFile, strerror(errno));
    yyparse();
    fclose(yyin);

    Device = Options.BootDevice;
    if (!Options.Boot.Default)
	Options.Boot.Default = BootRecords->Label;
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *	Add a New Boot Record
     */

int AddBootRecord(struct BootRecord *record)
{
    struct BootRecord **p, *new;

    for (p = &BootRecords; *p; p = &(*p)->Next)
	if (EqualStrings((*p)->Label, record->Label) ||
	    EqualStrings((*p)->Alias, record->Label) ||
	    EqualStrings((*p)->Label, record->Alias) ||
	    EqualStrings((*p)->Alias, record->Alias))
	    return(0);
    if (!(new = malloc(sizeof(struct BootRecord))))
	Die("No memory\n");
    *new = *record;
    *p = new;
    return(1);
}


    /*
     *	Create the Map File
     */

static void CreateMapFile(void)
{
    struct BootRecord *record, *defrecord = NULL;
    int ivecsize = 0, rvecsize = 0, i;

    if ((MapFileFH = open(MapFile, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) ==
	-1)
	Die("open %s: %s\n", MapFile, strerror(errno));

    /* File Identification */
    WriteTagData(TAG_LILO, NULL, 0);

    /* Boot Options */
    WriteTagData(TAG_HEADER, NULL, 0);
    if (Options.Boot.Default)
	WriteTagString(TAG_DEFAULT, Options.Boot.Default);
    if (Options.Boot.Auto)
	WriteTagData(TAG_AUTO, Options.Boot.Auto, 4);
    if (Options.Boot.TimeOut)
	WriteTagData(TAG_TIME_OUT, Options.Boot.TimeOut, 4);
    if (Options.Boot.Aux)
	WriteTagData(TAG_AUX, Options.Boot.Aux, 4);
    if (Options.Boot.Baud)
	WriteTagData(TAG_BAUD, Options.Boot.Baud, 4);
    if (Options.Boot.MasterPassword)
	WriteTagString(TAG_MASTER_PASSWORD, Options.Boot.MasterPassword);
    if (Options.Boot.Debug)
	WriteTagData(TAG_DEBUG, Options.Boot.Debug, 4);
    if (Options.Boot.Prompt)
	WriteTagString(TAG_PROMPT, Options.Boot.Prompt);
    WriteTagData(TAG_HEADER_END, NULL, 0);

    /* Boot Records */
    puts("Boot records:");
    for (record = BootRecords; record; record = record->Next) {
	if (EqualStrings(Options.Boot.Default, record->Label) ||
	    EqualStrings(Options.Boot.Default, record->Alias)) {
	    defrecord = record;
	    if (record->Password)
		Die("Default boot record `%s' must not have a password\n",
		    record->Label);
	}
	ivecsize = 0;
	rvecsize = 0;
	if (!record->OSType || *record->OSType != BOS_AMIGA) {
	    int numblocks;

	    if ((record->ImageVector = CreateVector(record->ImagePath,
						    &numblocks)))
		ivecsize = (numblocks+1)*sizeof(u_long);
	    else if (defrecord == record)
		Die("Kernel image `%s' for default boot record `%s' does not exist\n",
		    record->ImagePath, record->Label);
	    if (record->RamdiskPath &&
		(record->RamdiskVector = CreateVector(record->RamdiskPath,
						      &numblocks)))
		    rvecsize = (numblocks+1)*sizeof(u_long);
	}
	if (defrecord == record)
	    printf("\e[1m");
	printf("    %-32.32s (", record->Label);
	if (record->OSType && *record->OSType == BOS_AMIGA)
	    puts("ROM image)");
	else if (record->ImageVector)
	    printf("%ld bytes)\n", record->ImageVector[0]);
	else
	    puts("not available)");
	if (defrecord == record)
	    printf("\e[0m");

	WriteTagString(TAG_BOOT_RECORD, record->Label);
	if (record->Alias)
	    WriteTagString(TAG_ALIAS, record->Alias);
	if (record->OSType)
	    WriteTagData(TAG_OS_TYPE, record->OSType, 4);
	if (record->ImagePath)
	    WriteTagString(TAG_IMAGE_PATH, record->ImagePath);
	if (record->ImageVector)
	    WriteTagData(TAG_IMAGE_VECTOR, record->ImageVector, ivecsize);
	if (record->Args)
	    WriteTagString(TAG_ARGUMENTS, record->Args);
	if (record->Password)
	    WriteTagString(TAG_PASSWORD, record->Password);
	if (record->ChipSize)
	    WriteTagData(TAG_CHIP_RAM_SIZE, record->ChipSize, 4);
	for (i = 0; i < arraysize(record->FastChunks); i++)
	    if (record->FastChunks[i])
		WriteTagData(TAG_FAST_RAM_CHUNK, record->FastChunks[i], 8);
	if (record->Model)
	    WriteTagData(TAG_MODEL, record->Model, 4);
	if (record->RamdiskPath)
	    WriteTagString(TAG_RAMDISK_PATH, record->RamdiskPath);
	if (record->RamdiskVector)
	    WriteTagData(TAG_RAMDISK_VECTOR, record->RamdiskVector, rvecsize);
	WriteTagData(TAG_BOOT_RECORD_END, NULL, 0);
    }
    if (!defrecord)
	Die("Default boot record `%s' does not exist\n", Options.Boot.Default);

    /* End of File */
    WriteTagData(TAG_EOF, NULL, 0);
    close(MapFileFH);
}


    /*
     *	Write a Tag Record to the Map File
     */

static void WriteTagData(u_long tag, void *data, u_long size)
{
    struct TagRecord tr;

    tr.Tag = tag;
    tr.Size = size;
    if (write(MapFileFH, &tr, sizeof(tr)) != sizeof(tr))
	Die("write %s: %s\n", MapFile, strerror(errno));
    if (size) {
	if (write(MapFileFH, data, size) != size)
	    Die("write %s: %s\n", MapFile, strerror(errno));
    }
}


static void WriteTagString(u_long tag, char *s)
{
    struct TagRecord tr;

    tr.Tag = tag;
    tr.Size = (strlen(s)+4)&~3;
    if (write(MapFileFH, &tr, sizeof(tr)) != sizeof(tr))
	Die("write %s: %s\n", MapFile, strerror(errno));
    if (write(MapFileFH, s, tr.Size) != tr.Size)
	Die("write %s: %s\n", MapFile, strerror(errno));
}


    /*
     *	Calculate the Boot Block Checksum
     */

static u_long CalcChecksum(void)
{
    u_long _res;
    struct BootBlock *bb = &BootBlock;

    __asm __volatile ("1: addl %0@+,%1;"
		      "   addxl %5,%1;"
		      "   dbf %4,1b;"
		      : "=a" (bb), "=d" (_res)
		      : "0" (bb), "1" (0), "d" (255), "d" (0));
    return(~_res);
}


    /*
     *	Get the Geometry for the Partition a file belongs to
     */

static void Geometry(char *name, dev_t *device, u_long *start, u_long *blksize)
{
    struct hd_geometry geo;
    char devname[PATH_MAX+1];
    int fh, found = 0;

    if (stat(name, &info) == -1)
	Die("stat %s: %s\n", name, strerror(errno));
    if (!S_ISREG(info.st_mode))
	Die("%s is not a regular file\n", name);
    *device = info.st_dev;
    *blksize = info.st_blksize;

    if (!(dir = opendir("/dev")))
	Die("opendir /dev: %s\n", strerror(errno));
    while ((dirent = readdir(dir))) {
	sprintf(devname, "/dev/%s", dirent->d_name);
	if ((stat(devname, &info) != -1) && S_ISBLK(info.st_mode) &&
	    (info.st_rdev == *device)) {
	    if ((fh = open(devname, O_RDONLY)) == -1)
		Die("open %s: %s\n", devname, strerror(errno));
	    if (ioctl(fh, HDIO_GETGEO, &geo) == -1)
		Die("ioctl %s: %s\n", devname, strerror(errno));
	    close(fh);
	    found = 1;
	    break;
	}
    }
    closedir(dir);
    if (found)
	*start = 512*geo.start;
    else
	Die("Can't find special device entry for file `%s'\n");
}


    /*
     *	Create a Vector for a File
     */

static u_long *CreateVector(char *name, int *numblocks)
{
    int fh, size, i;
    u_long *vector = NULL;
    u_long offset, start, blksize;
    dev_t device;
    int relblksize;
    static u_int first = 1;
    static dev_t altdev;

    if (!name || ((fh = open(name, O_RDONLY)) == -1))
	return(NULL);
    if ((size = lseek(fh, 0, SEEK_END)) == -1)
	Die("lseek %s: %s\n", name, strerror(errno));
    *numblocks = (size+LILO_BLOCKSIZE-1)/LILO_BLOCKSIZE;
    if (!(vector = malloc((*numblocks+1)*sizeof(u_long))))
	Die("No memory\n");
    vector[0] = size;

    Geometry(name, &device, &start, &blksize);

    if (!Options.AltDeviceName) {
	if (MAJOR(device) != MAJOR(BootDevice))
	    Die("File `%s' must reside on physical device `%s'\n", name,
		Device);
	switch (MAJOR(BootDevice)) {
	    case FLOPPY_MAJOR:
		if (MINOR(device) != MINOR(BootDevice))
		    Die("File `%s' must reside on physical floppy device "
			"`%s'\n", name, Device);
		break;
	    case HD_MAJOR:
		if ((MINOR(device) & 0xc0) != (MINOR(BootDevice) & 0xc0))
		    Die("File `%s' must reside on physical IDE device `%s'\n",
			name, Device);
		break;
	    case SCSI_DISK_MAJOR:
		if ((MINOR(device) & 0xf0) != (MINOR(BootDevice) & 0xf0))
		    Die("File `%s' must reside on physical SCSI device `%s'\n",
			name, Device);
		break;
	    default:
		Die("Major device %d not supported (yet)\n", MAJOR(device));
		break;
	}
    } else if (first) {
	altdev = device;
	printf("Assuming `%s' resides on `%s:%ld'\n", name,
	       Options.AltDeviceName, *Options.AltDeviceUnit);
	first = 0;
    } else {
	if (MAJOR(device) != MAJOR(altdev))
	    Die("File `%s' must reside on `%s:%ld'\n", name,
		Options.AltDeviceName, Options.AltDeviceUnit);
	switch (MAJOR(altdev)) {
	    case FLOPPY_MAJOR:
		if (MINOR(device) != MINOR(altdev))
		    Die("File `%s' must reside on physical floppy device "
			"`%s:%ld'\n",
			name, Options.AltDeviceName, Options.AltDeviceUnit);
		break;
	    case HD_MAJOR:
		if ((MINOR(device) & 0xc0) != (MINOR(altdev) & 0xc0))
		    Die("File `%s' must reside on physical IDE device "
			"`%s:%ld'\n", name, Options.AltDeviceName,
			Options.AltDeviceUnit);
		break;
	    case SCSI_DISK_MAJOR:
		if ((MINOR(device) & 0xf0) != (MINOR(altdev) & 0xf0))
		    Die("File `%s' must reside on physical SCSI device "
			"`%s:%ld'\n", name, Options.AltDeviceName,
			Options.AltDeviceUnit);
		break;
	    default:
		Die("Major device %d not supported (yet)\n", MAJOR(device));
		break;
	}
    }

    if (blksize % LILO_BLOCKSIZE)
	Die("File `%s' resides on a device with an unsupported block size (%d "
	    "bytes)\n", name, blksize);

    relblksize = blksize/LILO_BLOCKSIZE;

    for (i = 0; i < *numblocks; i++) {
	if (!(i % relblksize)) {
	    offset = i/relblksize;
	    if (ioctl(fh, FIBMAP, &offset) == -1)
		Die("ioctl FIBMAP %s: %s\n", name, strerror(errno));
	    if (offset)
		offset = start+offset*blksize;
	} else if ((offset = vector[i]))
	    offset += LILO_BLOCKSIZE;
	vector[i+1] = offset;
    }
    close(fh);
    return(vector);
}


    /*
     *	Patch the Loader with the Map Vector
     */

static void PatchLoader(void)
{
    int i;
    u_long pattern[] = {
	LILO_ID, LILO_MAPVECTORID
    };
    u_long *data = NULL;
    u_long maxsize;

    for (i = 0; i < LoaderSize-sizeof(pattern)+1; i++)
	if (!strncmp(&LoaderData[i], (char *)pattern, sizeof(pattern))) {
	    data = (u_long *)&LoaderData[i];
	    break;
	}
    if (!data)
	Die("Couldn't find magic key in loader template\n");
    maxsize = data[2];
    if (MapNumBlocks > maxsize-1)
	Die("Map vector is too large for loader (%d entries too much)\n",
	    MapNumBlocks-maxsize-1);
    if (Verbose)
	printf("%d entries for map vector blocks, %lu entries free\n",
	       MapNumBlocks, maxsize-1-MapNumBlocks);
    memcpy(&data[3], MapVector, (MapNumBlocks+1)*sizeof(u_long));
}


    /*
     *	Create the Boot Block
     */

static void CreateBootBlock(void)
{
    u_long maxsize, size;

    size = _BootCodeEnd-_BootCodeStart;
    if (size > sizeof(BootBlock.Data))
	Die("%d bytes too much boot code\n", size-sizeof(BootBlock.Data));
    if (Verbose)
	printf("%lu bytes boot code, %lu bytes free\n", size,
	       sizeof(BootBlock.Data)-size);
    BootBlock.ID = DosTypeID;
    if (BootBlock.ID == ID_BOOT_DISK)
	BootBlock.DosBlock = 0;
    memcpy(BootBlock.Data, _BootCodeStart, size);

    maxsize = (_LoaderVectorEnd-_LoaderVectorStart)/sizeof(u_long);
    if (LoaderNumBlocks > maxsize-1)
	Die("Loader vector is too large for boot block (%d entries too much)\n",
	     LoaderNumBlocks-maxsize-1);
    if (Verbose)
	printf("%d entries for loader vector blocks, %lu entries free\n",
	       LoaderNumBlocks, maxsize-1-LoaderNumBlocks);
    memcpy((void *)((u_long)BootBlock.Data+(u_long)_LoaderVectorStart-
		    (u_long)_BootCodeStart), LoaderVector,
	   (LoaderNumBlocks+1)*sizeof(u_long));
    if (Options.AltDeviceName) {
	strcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceName-
			(u_long)_BootCodeStart), Options.AltDeviceName);
	memcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceUnit-
			(u_long)_BootCodeStart), &Options.AltDeviceUnit,
			     sizeof(Options.AltDeviceUnit));
    }
    BootBlock.LiloID = LILO_ID;
}


    /*
     *	Set the DosType of the Boot Block
     */

static void SetDosType(void)
{
    int i;

    for (i = 0; DosTypeEntries[i].ID; i++)
	if (!strcmp(DosType, DosTypeEntries[i].ASCII))
	    break;
    if (DosTypeEntries[i].ID)
	DosTypeID = DosTypeEntries[i].ID;
    else {
	printf("Unknown DosType `%s'\n", DosType);
	printf("Valid DosTypes are:\n");
	for (i = 0; DosTypeEntries[i].ID; i++)
	    printf(" %s", DosTypeEntries[i].ASCII);
	Die("\n");
    }
    BootBlock.ID = DosTypeID;
}


    /*
     *	Fix the Checksum of the Boot Block
     */

static void FixChecksum(void)
{
    BootBlock.Checksum = 0;
    BootBlock.Checksum = CalcChecksum();
}


    /*
     *	Print the Usage Template and Exit
     */

static void Usage(void)
{
    Die("Usage: %s [options]\n\n"
	"Valid options are:\n"
	"   -h, --help                : Display this usage information\n"
	"   -v, --verbose             : Enable verbose mode\n"
	"   -f, --force               : Force installation over an unknown "
	"boot block\n"
	"   -t, --dostype <dostype>   : Dostype for the boot block\n"
	"   -d, --device <device>     : Block special device for restore/save/"
	"checksum\n"
	"   -r, --restore-from <file> : File to restore the boot block from\n"
	"   -s, --save-to <file>      : File to save the boot block to\n"
	"   -c, --checksum            : Fix the boot block checksum\n"
	"   -w, --root <prefix>       : Use another root directory\n\n",
	ProgramName);
}


    /*
     *	Create a Complete File Name
     */

static char *CreateFileName(const char *name)
{
    int size;
    char *s;

    size = strlen(name)+1;
    if (Root)
	size += strlen(Root);
    if (!(s = malloc(size)))
	Die("No memory\n");
    if (Root) {
	strcpy(s, Root);
	strcat(s, name);
    } else
	strcpy(s, name);
    return(s);
}


    /*
     *	Main Routine
     */

int main(int argc, char *argv[])
{
    int i;

    ProgramName = argv[0];
    argc--;
    argv++;
    puts(VERSION "\n");

    while (argc) {
	if (!strcmp(argv[0], "-h") || !strcmp(argv[0], "--help"))
	    Usage();
	if (!strcmp(argv[0], "-v") || !strcmp(argv[0], "--verbose"))
	    Verbose = 1;
	else if (!strcmp(argv[0], "-f") || !strcmp(argv[0], "--force"))
	    Force = 1;
	else if (!strcmp(argv[0], "-d") || !strcmp(argv[0], "--device"))
	    if (--argc && !Device) {
		Device = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-r") || !strcmp(argv[0], "--restore-from"))
	    if (--argc && !RestoreBootBlock) {
		RestoreBootBlock = argv[1];
		Install = 0;
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-s") || !strcmp(argv[0], "--save-to"))
	    if (--argc && !SaveBootBlock) {
		SaveBootBlock = argv[1];
		Install = 0;
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-t") || !strcmp(argv[0], "--dostype"))
	    if (--argc && !DosType) {
		DosType = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-c") || !strcmp(argv[0], "--checksum")) {
	    Install = 0;
	    DoChecksum = 1;
	} else if (!strcmp(argv[0], "-w") || !strcmp(argv[0], "--root"))
	    if (--argc && !Root) {
		Root = argv[1];
		argv++;
	    } else
		Usage();
	else
	    Usage();
	argc--;
	argv++;
    }

    if (Device && Install && !DoChecksum)
	Die("You can specify a device only with -s, -r or -c\n");
    if (!Device && !Install)
	Die("You must specify a device if you use -s, -r or -c\n");

    ConfigFile = CreateFileName(LILO_CONFIGFILE);
    MapFile = CreateFileName(LILO_MAPFILE);
    LoaderTemplate = CreateFileName(LILO_LOADERTEMPLATE);
    LoaderFile = CreateFileName(LILO_LOADERFILE);

    if (Install)
	ReadConfigFile();

    if (stat(Device, &info) == -1)
	Die("stat %s: %s\n", Device, strerror(errno));
    if (!S_ISBLK(info.st_mode))
	Die("%s is not a block special device\n", Device);
    BootDevice = info.st_rdev;

    ReadBootBlock(Device);
    printf("%s: ", Device);
    DosTypeID = BootBlock.ID;
    for (i = 0; DosTypeEntries[i].ID; i++)
	if (BootBlock.ID == DosTypeEntries[i].ID)
	    break;
    if (DosTypeEntries[i].ID)
	printf("%s", DosTypeEntries[i].Description);
    else {
	printf("Type 0x%08lx", BootBlock.ID);
	if (!Force) {
	    printf("\n");
	    Die("Use option -f to force continue\n");
	}
	DosTypeID = ID_BOOT_DISK;
    }
    printf("%s, Boot block checksum is %s\n",
	   BootBlock.LiloID == LILO_ID ? " (created by Lilo)" : "",
	   CalcChecksum() ? "Incorrect" : "OK");

    if (SaveBootBlock) {
	WriteFile(SaveBootBlock, &BootBlock, sizeof(BootBlock));
	printf("Boot block saved to file `%s'\n", SaveBootBlock);
    }
    if (RestoreBootBlock) {
	ReadBootBlock(RestoreBootBlock);
	printf("Boot block restored from file `%s'\n", RestoreBootBlock);
    }
    if (DosType)
	SetDosType();
    if (Install) {
	CreateMapFile();
	LoaderSize = ReadFile(LoaderTemplate, (void **)&LoaderData);
	if (!(MapVector = CreateVector(MapFile, &MapNumBlocks)))
	    Die("open %s: %s\n", MapFile, strerror(errno));
	if (Verbose)
	    printf("Map vector has %lu bytes\n", MapVector[0]);
	PatchLoader();
	WriteFile(LoaderFile, LoaderData, LoaderSize);
	if (!(LoaderVector = CreateVector(LoaderFile, &LoaderNumBlocks)))
	    Die("open %s: %s\n", LoaderFile, strerror(errno));
	if (Verbose)
	    printf("Loader vector has %lu bytes\n", LoaderVector[0]);
	CreateBootBlock();
    }
    if (Install || DoChecksum)
	FixChecksum();
    if (RestoreBootBlock || Install || DoChecksum) {
	WriteBootBlock(Device);
	printf("Boot block written to device `%s'\n", Device);
    }
    exit(0);
}
