/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *		       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *  --------------------------------------------------------------------------
 *
 *  The original idea for this program comes from the Linux distribution:
 *
 *	LILO - Generic Boot Loader for Linux ("LInux LOader")
 *	       by Werner Almesberger
 *
 *  --------------------------------------------------------------------------
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file README.legal in the main directory of the
 *  Linux/m68k distribution for more details.
 *
 *  --------------------------------------------------------------------------
 *
 *  WARNING: Because this code must be PC-relative, it's compiled with the
 *	     -fPIC option and converted with my `fixpic' Perl script.
 *	     As a consequence we must not use pointers that are initialized at
 *	     compile time but we must explicitly initialize them at run time.
 *	     Look at the initialization of NewScreen.DefaultTitle and
 *	     NewWindow.Title for an example.
 *
 *  NOTE:    This applies to linuxboot.c too!
 */


#define V36_COMPATIBILITY	/* Make it work with the A3000 Beta ROM */


#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <linux/a.out.h>
#include <linux/elf.h>
#include <linux/linkage.h>
#include <asm/bootinfo.h>
#include <asm/amigatypes.h>
#include <asm/amigahw.h>

#include "loader.h"
#include "bootlib.h"
#include "lilo.h"


    /*
     *	Program Entry
     *
     *	Pass the Pointer to the LiloTable (in d0) to Main
     */

asm(".text\n"
ALIGN_STR "\n"
".globl " SYMBOL_NAME_STR(_start) ";\n"
SYMBOL_NAME_STR(_start) ":
	movel %d0,%sp@-
	jbsr  (" SYMBOL_NAME_STR(Main) ")
	addql #4,%sp
	rts
");


    /*
     *	Lilo Alert
     */

#define AN_LILO		(AT_DeadEnd|0x4c000000)	/* `L' */
#define AO_LiloMap	(0x004d4150)		/* `MAP': Corrupt Map File */


    /*
     *	Console Echo Modes
     */

#define ECHO_NONE	(0)
#define ECHO_NORMAL	(1)
#define ECHO_DOT	(2)


    /*
     *	Custom Hardware Bases
     *
     *	Make sure these are the physical and not the virtual addresses
     */

#undef custom
#define custom		((*(volatile struct CUSTOM *)(CUSTOM_PHYSADDR)))
#undef ciaa
#define ciaa		((*(volatile struct CIA *)(CIAA_PHYSADDR)))


    /*
     *	Boot Library Functions
     */

static struct LiloTable *LiloTable;


    /*
     *	Library Bases
     */

#ifdef V36_COMPATIBILITY
#define LIB_VERS	(36)
#else /* V36_COMPATIBILITY */
#define LIB_VERS	(37)
#endif /* V36_COMPATIBILITY */

struct ExecBase *SysBase;
struct ExpansionBase *ExpansionBase;
struct GfxBase *GfxBase;
static struct IntuitionBase *IntuitionBase;
static struct InputBase *InputBase;
#ifdef V36_COMPATIBILITY
static struct KeyboardBase *KeyboardBase;
#endif /* V36_COMPATIBILITY */


    /*
     *	Console Stuff
     */

static u_char Title[] = "Amiga Linux/m68k Loader";

static struct NewScreen NewScreen = {
    0, 0, 0, 0, 2, 0, 1, HIRES, CUSTOMSCREEN, NULL, NULL, NULL, NULL
};
static struct NewWindow NewWindow = {
    0, 0, 0, 0, 2, 3, 0, WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE, NULL, NULL,
    NULL, NULL, NULL, 0, 0, 0, 0, CUSTOMSCREEN
};
static struct Screen *Screen;
static struct Window *Window;
static struct MsgPort *ConsolePort;
static struct IOStdReq *ConsoleRequest = NULL;
static struct MsgPort *TimerPort;
static struct timerequest *TimerRequest;
static struct MsgPort *InputPort;
static struct IOStdReq *InputRequest;
#ifdef V36_COMPATIBILITY
static struct MsgPort *KeyboardPort;
static struct IOStdReq *KeyboardRequest;
static u_char KeyboardMatrix[13];
#endif /* V36_COMPATIBILITY */
static u_long TimeOutSignalMask = 0;
static char FmtBuffer[256], InBuffer[CL_SIZE];
static u_long ConsoleOpened = FALSE;
static void *OldDebug;


    /*
     *	Map Data
     */

#define MAXVECTORSIZE	(1024)

struct FileVectorData {
    u_long ID[2];
    u_long MaxVectorSize;
    u_long Vector[MAXVECTORSIZE];
};

static struct FileVectorData MapVectorData = {
    { LILO_ID, LILO_MAPVECTORID }, MAXVECTORSIZE
};

static u_char *MapData = NULL;
static u_long MapSize = 0, MapOffset = 0;
static struct BootOptions *BootOptions = NULL;
static struct BootRecord *BootRecords = NULL;


static u_short MasterMode = FALSE;
static u_short Auto = FALSE;
static u_short Aux = FALSE;
static u_short Debug = FALSE;
static char *Prompt = NULL;
static u_short AutoBoot = TRUE;


    /*
     *	Linux/m68k Bootstrap Stuff
     */

static struct BootRecord *BootRecord;


    /*
     *	Function Prototypes
     */

u_long Main(struct LiloTable *lilotable);
static u_long GetDebugTag(void);
static char *GetTagName(u_long tag);
static struct TagRecord *GetNextTag(void);
static void UnexpectedTag(struct TagRecord *tr);
static int EqualStrings(const char *s1, const char *s2);
static void ReadMapData(void);
static void FreeMapData(void);
static struct BootRecord *FindBootRecord(const char *name);
#ifdef V36_COMPATIBILITY
static u_short PeekQualifier36(void);
#endif /* V36_COMPATIBILITY */
static u_long ProbeConsole();
static void OpenConsole(void);
static void CloseConsole();
static long ReadLine(u_long mode);
static void SetTimeOut(u_long seconds);
static int TimedOut(void);
static void ClearTimeOut(void);
static struct BootRecord *GetDefaultBootOS(void);
static struct BootRecord *GetBootOS(void);
static void CreateCommandLine(void);
static u_long BootAmigaOS(void);


    /*
     *	Main Loader Routine
     *
     *	This routine should return FALSE in case of an error (which will cause
     *	a reboot), and TRUE if the user wants to boot AmigaOS.
     *	In all other cases it won't return :-)
     */

u_long Main(struct LiloTable *lilotable)
{
    u_long res = FALSE;
    u_long type;
    u_short interactive = FALSE;

    /* Boot Library */
    LiloTable = lilotable;

    /* Open Libraries */
    SysBase = *(struct ExecBase **)4;
    if (!(ExpansionBase =
	  (struct ExpansionBase *)OpenLibrary("expansion.library", LIB_VERS)))
	Alert(AN_LILO|AG_OpenLib|AO_ExpansionLib);
    if (!(GfxBase = (struct GfxBase *)OpenLibrary("graphics.library",
						  LIB_VERS)))
	Alert(AN_LILO|AG_OpenLib|AO_GraphicsLib);

    /* Open Timer Device */
    if (!(TimerPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(TimerRequest = CreateIORequest(TimerPort,
					 sizeof(struct timerequest))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("timer.device", UNIT_VBLANK,
		   (struct IORequest *)TimerRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_TimerDev);

    /* Open Input Device */
    if (!(InputPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(InputRequest = CreateIORequest(InputPort, sizeof(struct IOStdReq))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("input.device", 0, (struct IORequest *)InputRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_Unknown);
    InputBase = (struct InputBase *)InputRequest->io_Device;

    /* Configuration */
    ReadMapData();
    if (!Auto || ProbeConsole()) {
	OpenConsole();
	interactive = TRUE;
    }

    do {
	BootRecord = interactive ? GetBootOS() : GetDefaultBootOS();
	if (!BootRecord) {
	    Puts("\nPanic: No boot Operating System!\n");
	    break;
	}
	CreateCommandLine();
	type = BootRecord->OSType ? *BootRecord->OSType : BOS_LINUX;
	switch (type) {
	    case BOS_AMIGA:
		Printf("\nBooting AmigaOS (%s)...\n\n", BootRecord->Label);
		res = BootAmigaOS();
		break;
	    case BOS_LINUX:
		Printf("\nBooting Linux/m68k (%s)...\n\n", BootRecord->Label);
		res = BootLinux(TAG_IMAGE_VECTOR,
				BootRecord->RamdiskVector ? TAG_RAMDISK_VECTOR :
							    0,
				BootRecord->Args, Debug, 0);
		break;
	    case BOS_NETBSD:
		Printf("\nBooting NetBSD-Amiga (%s)...\n\n", BootRecord->Label);
		Puts("Not yet implemented :-)\n");
		res = FALSE;
		break;
	}
	if (res)
	    break;
	if (BootRecord->ImageVector)
	    Puts("\nCouldn't boot desired Operating System\n");
	else
	    Puts("\nBoot Operating System is not available\n");
    } while (interactive);

    CloseConsole();
    FreeMapData();

    /* Close Input Device */
    CloseDevice((struct IORequest *)InputRequest);
    DeleteIORequest(InputRequest);
    DeleteMsgPort(InputPort);

    /* Close Timer Device */
    CloseDevice((struct IORequest *)TimerRequest);
    DeleteIORequest(TimerRequest);
    DeleteMsgPort(TimerPort);

    /* Close Libraries */
    CloseLibrary((struct Library*)GfxBase);
    CloseLibrary((struct Library *)ExpansionBase);

    return(res);
}


    /*
     *	Get the Value of the Debug Tag (if it exists)
     */

static u_long GetDebugTag(void)
{
    struct TagRecord *tr;
    u_long offset = 0;
    u_long debug = FALSE;

    while (offset+sizeof(struct TagRecord)+sizeof(u_long) <= MapSize) {
	tr = (struct TagRecord *)&MapData[offset];
	if (tr->Tag == TAG_DEBUG) {
	    debug = tr->Data[0];
	    break;
	}
	offset += sizeof(struct TagRecord)+tr->Size;
    }
    return(debug);
}


    /*
     *	Get the Name for a Tag
     */

static char *GetTagName(u_long tag)
{
    char *name = NULL;

    switch (tag) {
	case TAG_LILO:
	    name = "TAG_LILO";
	    break;
	case TAG_HEADER:
	    name = "TAG_HEADER";
	    break;
	case TAG_HEADER_END:
	    name = "TAG_HEADER_END";
	    break;
	case TAG_DEFAULT:
	    name = "TAG_DEFAULT";
	    break;
	case TAG_AUTO:
	    name = "TAG_AUTO";
	    break;
	case TAG_TIME_OUT:
	    name = "TAG_TIME_OUT";
	    break;
	case TAG_AUX:
	    name = "TAG_AUX";
	    break;
	case TAG_BAUD:
	    name = "TAG_BAUD";
	    break;
	case TAG_MASTER_PASSWORD:
	    name = "TAG_MASTER_PASSWORD";
	    break;
	case TAG_DEBUG:
	    name = "TAG_DEBUG";
	    break;
	case TAG_PROMPT:
	    name = "TAG_PROMPT";
	    break;
	case TAG_BOOT_RECORD:
	    name = "TAG_BOOT_RECORD";
	    break;
	case TAG_BOOT_RECORD_END:
	    name = "TAG_BOOT_RECORD_END";
	    break;
	case TAG_ALIAS:
	    name = "TAG_ALIAS";
	    break;
	case TAG_OS_TYPE:
	    name = "TAG_OS_TYPE";
	    break;
	case TAG_IMAGE_PATH:
	    name = "TAG_IMAGE_PATH";
	    break;
	case TAG_IMAGE_VECTOR:
	    name = "TAG_IMAGE_VECTOR";
	    break;
	case TAG_ARGUMENTS:
	    name = "TAG_ARGUMENTS";
	    break;
	case TAG_PASSWORD:
	    name = "TAG_PASSWORD";
	    break;
	case TAG_CHIP_RAM_SIZE:
	    name = "TAG_CHIP_RAM_SIZE";
	    break;
	case TAG_FAST_RAM_CHUNK:
	    name = "TAG_FAST_RAM_CHUNK";
	    break;
	case TAG_MODEL:
	    name = "TAG_MODEL";
	    break;
	case TAG_RAMDISK_PATH:
	    name = "TAG_RAMDISK_PATH";
	    break;
	case TAG_RAMDISK_VECTOR:
	    name = "TAG_RAMDISK_VECTOR";
	    break;
	case TAG_EOF:
	    name = "TAG_EOF";
	    break;
    }
    return(name);
}


    /*
     *	Get the Next Tag from the Map Data
     */

static struct TagRecord *GetNextTag(void)
{
    static struct TagRecord *tr = NULL;
    char *name;

    if (tr)
	MapOffset += sizeof(struct TagRecord)+tr->Size;
    if (MapOffset+sizeof(struct TagRecord) > MapSize)
	Alert(AN_LILO|AO_LiloMap);
    tr = (struct TagRecord *)&MapData[MapOffset];
    if (Debug) {
        name = GetTagName(tr->Tag);
	Printf("0x%08lx: 0x%08lx bytes (0x%08lx, `%s')\n", MapOffset, tr->Size,
	       tr->Tag, name ? name : "<UNKNOWN>");
    }
    return(tr);
}


    /*
     *	Unexpected Tag
     */

static void UnexpectedTag(struct TagRecord *tr)
{
    char *name = GetTagName(tr->Tag);

    if (name)
	Printf("Unexpected tag `%s' at offset 0x%08lx\n", name, MapOffset);
    else
	Printf("Unknown tag 0x%08lx at offset 0x%08lx\n", tr->Tag, MapOffset);
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *	Read the Map Data
     */

static void ReadMapData(void)
{
    struct TagRecord *tr;
    struct BootRecord *record;
    struct BootRecord **p = &BootRecords;
    u_short isdefault = FALSE;
    int i;

    Lilo_Open(MapVectorData.Vector);
    MapSize = MapVectorData.Vector[0];
    if (!(MapData = (u_char *)AllocMem(MapSize, MEMF_PUBLIC|MEMF_CLEAR)))
	Alert(AN_LILO|AG_NoMemory);
    Lilo_Read(MapData, MapSize);
    Lilo_Close();
    if ((Debug = GetDebugTag()))
	OpenConsole();

    /* File Identification */
    tr = GetNextTag();
    if (tr->Tag != TAG_LILO)
	Alert(AN_LILO|AO_LiloMap);

    /* Boot Options */
    tr = GetNextTag();
    if (tr->Tag != TAG_HEADER)
	Alert(AN_LILO|AO_LiloMap);
    if (!(BootOptions = (struct BootOptions *)AllocMem(sizeof(*BootOptions),
						       MEMF_PUBLIC|MEMF_CLEAR)))
	Alert(AN_LILO|AG_NoMemory);
    for (tr = GetNextTag(); tr->Tag != TAG_HEADER_END; tr = GetNextTag())
	switch (tr->Tag) {
	    case TAG_DEFAULT:
		BootOptions->Default = (char *)tr->Data;
		break;
	    case TAG_AUTO:
		BootOptions->Auto = (u_long *)tr->Data;
		break;
	    case TAG_TIME_OUT:
		BootOptions->TimeOut = (u_long *)tr->Data;
		break;
	    case TAG_AUX:
		BootOptions->Aux = (u_long *)tr->Data;
		break;
	    case TAG_BAUD:
		BootOptions->Baud = (u_long *)tr->Data;
		break;
	    case TAG_MASTER_PASSWORD:
		BootOptions->MasterPassword = (char *)tr->Data;
		break;
	    case TAG_DEBUG:
		BootOptions->Debug = (u_long *)tr->Data;
		break;
	    case TAG_PROMPT:
		BootOptions->Prompt = (char *)tr->Data;
		break;
	    default:
		UnexpectedTag(tr);
		break;;
	}

    if (BootOptions->Auto)
	Auto = *BootOptions->Auto;
    if (BootOptions->Aux)
	Aux = *BootOptions->Aux;
    if (BootOptions->Debug)
	Debug = *BootOptions->Debug;
    if (BootOptions->Baud) {
	u_long baud = *BootOptions->Baud;
	if (baud)
	    custom.serper = (5*SysBase->ex_EClockFrequency+baud/2)/baud-1;
    }
    if (!BootOptions->Default)
	isdefault = TRUE;
    Prompt = BootOptions->Prompt ? BootOptions->Prompt : "Boot image: ";

    if (Debug) {
	Printf("Map file contains %ld bytes of data\n\n", MapSize);
	Printf("Auto:            %s\n", Auto ? "Enabled" : "Disabled");
	Printf("Debug:           %s\n", Debug ? "Enabled" : "Disabled");
	Printf("Aux:             %s\n", Aux ? "Enabled" : "Disabled");
	if (BootOptions->Default)
	    Printf("Default Image:   `%s'\n", BootOptions->Default);
	if (BootOptions->TimeOut)
	    Printf("Time Out:        %ld\n", *BootOptions->TimeOut);
	if (BootOptions->MasterPassword)
	    Printf("Master Password: `%s'\n", BootOptions->MasterPassword);
	if (BootOptions->Baud)
	    Printf("Baud:            %ld\n", *BootOptions->Baud);
    }

    /* Boot Records */
    tr = GetNextTag();
    if (Debug)
	Puts("\nBoot records:\n");
    do {
	if (tr->Tag != TAG_BOOT_RECORD)
	    Alert(AN_LILO|AO_LiloMap);
	if (!(record = (struct BootRecord *)AllocMem(sizeof(*record),
						     MEMF_PUBLIC|MEMF_CLEAR)))
	    Alert(AN_LILO|AG_NoMemory);
	record->Label = (char *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_BOOT_RECORD_END;
	     tr = GetNextTag())
	    switch (tr->Tag) {
		case TAG_ALIAS:
		    record->Alias = (u_char *)tr->Data;
		    break;
		case TAG_OS_TYPE:
		    record->OSType = (u_long *)tr->Data;
		    break;
		case TAG_IMAGE_PATH:
		    record->ImagePath = (u_char *)tr->Data;
		    break;
		case TAG_IMAGE_VECTOR:
		    record->ImageVector = (u_long *)tr->Data;
		    break;
		case TAG_ARGUMENTS:
		    record->Args = (u_char *)tr->Data;
		    break;
		case TAG_PASSWORD:
		    record->Password = (u_char *)tr->Data;
		    break;
		case TAG_CHIP_RAM_SIZE:
		    record->ChipSize = (u_long *)tr->Data;
		    break;
		case TAG_FAST_RAM_CHUNK:
		    for (i = 0; i < arraysize(record->FastChunks); i++)
			if (!record->FastChunks[i])
			    break;
		    if (i < arraysize(record->FastChunks))
			record->FastChunks[i] = (u_long *)tr->Data;
		    else
			UnexpectedTag(tr);
		    break;
		case TAG_MODEL:
		    record->Model = (u_long *)tr->Data;
		    break;
		case TAG_RAMDISK_PATH:
		    record->RamdiskPath = (u_char *)tr->Data;
		    break;
		case TAG_RAMDISK_VECTOR:
		    record->RamdiskVector = (u_long *)tr->Data;
		    break;
		default:
		    UnexpectedTag(tr);
		    break;;
	    }

	if (EqualStrings(BootOptions->Default, record->Label) ||
	    EqualStrings(BootOptions->Default, record->Alias))
	    isdefault = TRUE;

	if (Debug) {
	    if (isdefault)
		Printf("\e[1m");
	    Printf("    %-32.32s (", record->Label);
	    if (record->OSType && *record->OSType == BOS_AMIGA)
		Puts("ROM image)");
	    else if (record->ImageVector)
		Printf("%ld bytes)\n", record->ImageVector[0]);
	    else
		Puts("not available)");
	    if (isdefault) {
		Printf("\e[0m");
		isdefault = FALSE;
	    }
	}

	*p = record;
	p = &record->Next;
	tr = GetNextTag();
    } while (tr->Tag != TAG_EOF);
}


    /*
     *	Free the Map Data
     */

static void FreeMapData(void)
{
    struct BootRecord *record, *next;

    for (record = BootRecords; record; record = next) {
	next = record->Next;
	FreeMem(record, sizeof(struct BootRecord));
    }
    if (BootOptions)
	FreeMem(BootOptions, sizeof(struct BootOptions));
    if (MapData)
	FreeMem(MapData, MapSize);
}


    /*
     *	Find a Boot Record by Name
     */

static struct BootRecord *FindBootRecord(const char *name)
{
    struct BootRecord *record;

    for (record = BootRecords; record; record = record->Next)
	if (EqualStrings(name, record->Label) ||
	    EqualStrings(name, record->Alias))
	    break;
    return(record);
}


#ifdef V36_COMPATIBILITY
    /*
     *	V36 (A3000 Beta ROM) compatible replacement for PeekQualifier()
     */

static u_short PeekQualifier36(void)
{
    u_short keys;

    /* Open Keyboard Device */

    if (!(KeyboardPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(KeyboardRequest = CreateIORequest(KeyboardPort,
					    sizeof(struct IOStdReq))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("keyboard.device", 0, (struct IORequest *)KeyboardRequest,
		   0))
	Alert(AN_LILO|AG_OpenDev|AO_KeyboardDev);
    KeyboardBase = (struct KeyboardBase *)KeyboardRequest->io_Device;

    KeyboardRequest->io_Command = KBD_READMATRIX;
    KeyboardRequest->io_Flags = IOF_QUICK;
    KeyboardRequest->io_Length = sizeof(KeyboardMatrix);
    KeyboardRequest->io_Data = KeyboardMatrix;
    DoIO((struct IORequest *)KeyboardRequest);

    keys = KeyboardMatrix[12];
    if (!(custom.potgor & 0x0100))
	keys |= IEQUALIFIER_MIDBUTTON;
    if (!(custom.potgor & 0x0400))
	keys |= IEQUALIFIER_RBUTTON;
    if (!(ciaa.pra & 0x40))
	keys |= IEQUALIFIER_LEFTBUTTON;

    /* Close Keyboard Device */

    CloseDevice((struct IORequest *)KeyboardRequest);
    DeleteIORequest(KeyboardRequest);
    DeleteMsgPort(KeyboardPort);

    return(keys);
}
#endif /* V36_COMPATIBILITY */


    /*
     *	Probe the Console
     */

static u_long ProbeConsole(void)
{
    u_long res;

#ifdef V36_COMPATIBILITY
    if (SysBase->Version >= 37)
#endif /* V36_COMPATIBILITY */
	res = PeekQualifier() ? TRUE : FALSE;
#ifdef V36_COMPATIBILITY
    else
	res = PeekQualifier36() ? TRUE : FALSE;
#endif /* V36_COMPATIBILITY */

    if (!res && Aux) {
	SetTimeOut(1);
	do
	    if (RawMayGetChar() != -1) {
		res = TRUE;
		while (RawMayGetChar() != -1);
		break;
	    }
	while (!TimedOut());
	ClearTimeOut();
    }
    return(res);
}


    /*
     *	Replacement for exec.library/Debug()
     *
     *	We don't want to meet a debugger when DEL is pressed on the terminal
     */

asm(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(NewDebug) ":
	rts
");

extern void NewDebug(void);


    /*
     *	Open the Console
     */

static void OpenConsole(void)
{
    if (!ConsoleOpened) {
	if (!(IntuitionBase =
	      (struct IntuitionBase *)OpenLibrary("intuition.library",
						  LIB_VERS)))
	    Alert(AN_LILO|AG_OpenLib|AO_Intuition);
	NewScreen.Width = GfxBase->NormalDisplayColumns;
	NewScreen.Height = GfxBase->NormalDisplayRows;
	NewScreen.DefaultTitle = Title;
	if (!(Screen = OpenScreen(&NewScreen)))
	    Alert(AN_LILO|AN_OpenScreen);
	NewWindow.Width = NewScreen.Width;
	NewWindow.Height = NewScreen.Height;
	NewWindow.Screen = Screen;
	NewWindow.Title = Title;
	if (!(Window = OpenWindow(&NewWindow)))
	    Alert(AN_LILO|AN_OpenWindow);
	if (!(ConsolePort = CreateMsgPort()))
	    Alert(AN_LILO|AG_NoSignal);
	if (!(ConsoleRequest = CreateIORequest(ConsolePort,
					       sizeof(struct IOStdReq))))
	    Alert(AN_LILO|AG_NoMemory);
	ConsoleRequest->io_Data = Window;
	if (OpenDevice("console.device", CONU_SNIPMAP,
		       (struct IORequest *)ConsoleRequest, CONFLAG_DEFAULT))
	    Alert(AN_LILO|AG_OpenDev|AO_ConsoleDev);
	if (Aux)
	    OldDebug = SetFunction((struct Library *)SysBase, LVODebug,
				   NewDebug);
	ConsoleOpened = TRUE;
	Puts(VERSION "\n\n");
    }
}


    /*
     *	Close the Console
     */

static void CloseConsole()
{
    if (ConsoleOpened) {
	if (Aux)
	    SetFunction((struct Library *)SysBase, LVODebug, OldDebug);
	CloseDevice((struct IORequest *)ConsoleRequest);
	DeleteIORequest(ConsoleRequest);
	DeleteMsgPort(ConsolePort);
	CloseWindow(Window);
	CloseScreen(Screen);
	CloseLibrary((struct Library*)IntuitionBase);
	ConsoleOpened = FALSE;
    }
}


    /*
     *	Print a String to the Console
     */

void Puts(char *str)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = -1;
	ConsoleRequest->io_Data = str;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux)
	    while (*str)
		RawPutChar(*str++);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
}


    /*
     *	Put a Character to the Console
     */

void PutChar(char c)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux)
	    RawPutChar(c);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
}


    /*
     *	Get a Character from the Console
     */

long GetChar(void)
{
    long res = -1;
    char c;

    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_READ;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	do
	    if (Aux && ((res = RawMayGetChar()) != -1))
		break;
	    else if (CheckIO((struct IORequest *)ConsoleRequest)) {
		res = c;
		break;
	    }
	while (MasterMode || Debug || !TimedOut());
	AbortIO((struct IORequest *)ConsoleRequest);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
    return(res);
}


    /*
     *	Format and Print a String to the Console
     */

asm(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(PutCharInBuf) ":
        movel   %a0,%sp@-
        lea     %pc@(" SYMBOL_NAME_STR(FmtBuffer) "+255),%a0
        cmpl    %a0,%a3
        jne     1f
        clrb    %a3@
        jra     2f
1:      moveb   %d0,%a3@+
2:      movel   %sp@+,%a0
        rts
");

extern void *PutCharInBuf(char c);

void Printf(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    RawDoFmt(fmt, args, (void (*)())PutCharInBuf, FmtBuffer);
    Puts(FmtBuffer);
    va_end(args);
}


    /*
     *	Read a Line from the Console
     */

static long ReadLine(u_long mode)
{
    long c;
    long length = 0;
    u_long eol = FALSE;

    while (!eol && ((c = GetChar()) != -1))
	switch (c) {
	    case '\r':
	    case '\n':
		InBuffer[length] = '\0';
		if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
		    PutChar('\n');
		eol = TRUE;
		break;
	    case '\b':
		if (length > 0) {
		    length--;
		    if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
			Puts("\b \b");
		}
		break;
	    case '\t':
		strcpy(InBuffer, "help");
		length = strlen("help");
		switch (mode) {
		    case ECHO_NORMAL:
			Puts("help\n");
			break;
		    case ECHO_DOT:
			Puts("++++\n");
			break;
		}
		eol = TRUE;
		break;
	    default:
		if (length < sizeof(InBuffer)-1) {
		    InBuffer[length++] = c;
		    switch (mode) {
			case ECHO_NORMAL:
			    PutChar(c);
			    break;
			case ECHO_DOT:
			    PutChar('+');
			    break;
		    }
		}
		break;
	}
    return(eol ? length : -1);
}


    /*
     *	Set the Time Out Value
     */

static void SetTimeOut(u_long seconds)
{
    if (!TimeOutSignalMask) {
	TimerRequest->io_Command = TR_ADDREQUEST;
	TimerRequest->tv_secs = seconds;
	TimerRequest->tv_micro = 0;
	SendIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 1<<TimerPort->mp_SigBit;
    }
}


    /*
     *	Check for a Time Out
     */

static int TimedOut(void)
{
    int res = FALSE;

    if (TimeOutSignalMask && CheckIO((struct IORequest *)TimerRequest))
	res = TRUE;

    return(res);
}


    /*
     *	Clear Time Out
     */

static void ClearTimeOut(void)
{
    if (TimeOutSignalMask) {
	AbortIO((struct IORequest *)TimerRequest);
	WaitIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 0;
    }
}


    /*
     *	Sleep for a Specified Period
     */

void Sleep(u_long micros)
{
    TimerRequest->io_Command = TR_ADDREQUEST;
    TimerRequest->io_Flags = IOF_QUICK;
    TimerRequest->tv_secs = micros/1000000;
    TimerRequest->tv_micro = micros%1000000;
    DoIO((struct IORequest *)TimerRequest);
}


    /*
     *	Get the Default Boot Operating System
     */

static struct BootRecord *GetDefaultBootOS(void)
{
    return(BootOptions->Default ? FindBootRecord(BootOptions->Default) : NULL);
}


    /*
     *	Ask the User for a Boot Operating System
     */

static struct BootRecord *GetBootOS(void)
{
    struct BootRecord *bootrecord = GetDefaultBootOS();
    struct BootRecord *record;
    u_long type;

    if (BootOptions->TimeOut ? *BootOptions->TimeOut : 20)
	SetTimeOut(*BootOptions->TimeOut);

    while (1) {
	Printf("\n%s", Prompt);
	if ((ReadLine(ECHO_NORMAL)) < 1)
	    break;
	if (!strcmp(InBuffer, "help") || !strcmp(InBuffer, "?")) {
	    Puts("\nValid commands: su, help, ?\n\nList of boot images:\n\n");
	    for (record = BootRecords; record; record = record->Next) {
		type = record->OSType ? *record->OSType : BOS_LINUX;
		if (record == bootrecord)
		    Puts("\e[1m");
		Printf("    %-31.31s (%s%s%s)\n", record->Label,
		       type == BOS_AMIGA ? "AmigaOS" :
		       type == BOS_LINUX ? "Linux/m68k" :
		       type == BOS_NETBSD ? "NetBSD-Amiga" : "Unknown",
		       record->Password ? ", Restricted" : "",
		       record->ImageVector || type == BOS_AMIGA ? "" : ", N/A");
		if (record == bootrecord)
		    Puts("\e[0m");
	    }
	} else if (!strcmp(InBuffer, "su")) {
	    if (MasterMode)
		continue;
	    if (BootOptions->MasterPassword) {
		Puts("\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		     strcmp(InBuffer, BootOptions->MasterPassword)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    MasterMode = TRUE;
	    Puts("\nEntering Master Mode.\n");
	} else {
	    if (!(record = FindBootRecord(InBuffer))) {
		Printf("\nNonexistent boot image `%s'.\n", InBuffer);
		continue;
	    }
	    type = record->OSType ? *record->OSType : BOS_LINUX;
	    if (type != BOS_AMIGA && !record->ImageVector) {
		Printf("\nBoot image `%s' is not available.\n", InBuffer);
		continue;
	    } else if (!MasterMode && record->Password) {
		Puts("\n\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		    strcmp(InBuffer, record->Password)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    bootrecord = record;
	    AutoBoot = FALSE;
	    break;
	}
    }
    ClearTimeOut();

    return(bootrecord);
}


    /*
     *	Create the Kernel Command Line
     */

static void CreateCommandLine(void)
{
    if (BootRecord->Args)
	strncpy(InBuffer, BootRecord->Args, CL_SIZE);
    else
	InBuffer[0] = '\0';
    strncat(InBuffer, " BOOT_IMAGE=", CL_SIZE);
    strncat(InBuffer, BootRecord->Label, CL_SIZE);
    InBuffer[CL_SIZE-1] = '\0';
    if (AutoBoot)
	strncat(InBuffer, " auto", CL_SIZE);
    if (MasterMode) {
	Printf("\n\nOld command line: %s\n\nNew command line: ", InBuffer);
	ReadLine(ECHO_NORMAL);
    }
    BootRecord->Args = InBuffer;
}


    /*
     *	Boot The Standard Amiga Operating System
     */

static u_long BootAmigaOS(void)
{
    return(TRUE);
}


    /*
     *	Functions needed by LinuxBoot
     *
     *	These functions are not complete implementations. They do only what's
     *	really necessary.
     */

static int OpenFiles = 0;

int Open(u_long key)
{
    if (OpenFiles++) {
	Puts("Internal error: Only one file can be open at the same time\n");
	return(-1);
    }
    if (key == TAG_IMAGE_VECTOR)
	Lilo_Open(BootRecord->ImageVector);
    else if (key == TAG_RAMDISK_VECTOR)
	Lilo_Open(BootRecord->RamdiskVector);
    else
	return(-1);
    return(1);
}


int Seek(int fd, int offset)
{
    Lilo_Seek(offset);
    return(offset);
}


int Read(int fd, char *buf, int count)
{
    Lilo_Read(buf, count);
    return(count);
}


void Close(int fd)
{
    OpenFiles--;
    Lilo_Close();
}


int FileSize(u_long key)
{
    if (key == TAG_IMAGE_VECTOR)
	return(BootRecord->ImageVector[0]);
    else if (key == TAG_RAMDISK_VECTOR)
	return(BootRecord->RamdiskVector[0]);
    return(-1);
}


    /*
     *	Modify the bootinfo, if necessary
     */

int ModifyBootInfo(struct bootinfo *bi)
{
    int i, j;

    if (Debug)
	Puts("Modifying bootinfo structure\n");

    if (BootRecord->ChipSize) {
	bi->bi_amiga.chip_size = *BootRecord->ChipSize;
	if (Debug)
	    Printf("Chip RAM: %10ld bytes\n", bi->bi_amiga.chip_size);
    }

    for (i = 0, j = 0; i < arraysize(BootRecord->FastChunks); i++)
	if (BootRecord->FastChunks[i]) {
	    bi->memory[j].addr = BootRecord->FastChunks[i][0];
	    bi->memory[j].size = BootRecord->FastChunks[i][1];
	    if (Debug)
		Printf("Fast RAM: %10ld bytes at 0x%08lx\n", bi->memory[j].addr,
		       bi->memory[j].size);
	    j++;
	}
    if (j > 0 && j != bi->num_memory)
	bi->num_memory = j;

    if (BootRecord->Model) {
	bi->bi_amiga.model = *BootRecord->Model;
	if (Debug)
	    Printf("Model   : #%ld\n", bi->bi_amiga.model);
    }

    return(TRUE);
}
